@@ -16,9 +16,10 @@ typedef websocketpp::client<websocketpp::config::asio_client> client;
 
 using namespace cadabra;
 
-ComputeThread::ComputeThread(int server_port)
+ComputeThread::ComputeThread(int server_port, std::string token, std::string ip_address)
 	: gui(0), docthread(0), connection_is_open(false), restarting_kernel(false), server_pid(0),
-	  server_stdout(0), server_stderr(0), forced_server_port(server_port)
+	  server_stdout(0), server_stderr(0), forced_server_port(server_port), forced_server_token(token),
+	  forced_server_ip_address(ip_address)
 	{
 	// The ComputeThread constructor (but _not_ the run() member!) is
 	// always run on the gui thread, so we can grab the gui thread id
@@ -79,7 +80,7 @@ void ComputeThread::try_connect()
 	                                  websocketpp::lib::placeholders::_2));
 
 	std::ostringstream uristr;
-	uristr << "ws://127.0.0.1:" << port;
+	uristr << "ws://" << forced_server_ip_address << ":" << port;
 	websocketpp::lib::error_code ec;
 	connection = wsclient.get_connection(uristr.str(), ec);
 	if (ec) {
@@ -129,6 +130,11 @@ void ComputeThread::terminate()
 		}
 	}
 
+bool ComputeThread::kernel_is_connected() const
+	{
+	return connection_is_open;
+	}
+
 void ComputeThread::all_cells_nonrunning()
 	{
 	for(auto it: running_cells) {
@@ -175,6 +181,7 @@ void ComputeThread::try_spawn_server()
 
 	if(forced_server_port!=0) {
 		port=forced_server_port;
+		authentication_token=forced_server_token;
 		return;
 		}
 
@@ -182,7 +189,15 @@ void ComputeThread::try_spawn_server()
 #if defined(_WIN32) || defined(_WIN64)
 	argv.push_back(cadabra::install_prefix()+"\\bin\\cadabra-server.exe");
 #else
-	argv.push_back("cadabra-server");
+	const char *appdir = getenv("APPDIR");
+	if(appdir) {
+		// std::cerr << "This is an AppImage, APPDIR = " << appdir << std::endl;
+		argv.push_back(std::string(appdir)+"/usr/bin/cadabra-server");
+		}
+	else {
+		// std::cerr << "Not an AppImage." << std::endl;
+		argv.push_back("cadabra-server");
+		}
 #endif
 	Glib::Pid pid;
 	std::string wd("");
@@ -201,10 +216,15 @@ void ComputeThread::try_spawn_server()
 
 		char buffer[100];
 		FILE *f = fdopen(server_stdout, "r");
-		if(fscanf(f, "%100s", buffer)!=1) {
+		if(fscanf(f, "%99s", buffer)!=1) {
 			throw std::logic_error("Failed to read port from server.");
 			}
 		port = atoi(buffer);
+		if(fscanf(f, "%99s", buffer)!=1) {
+			throw std::logic_error("Failed to read authentication token from server.");
+			}
+		authentication_token=std::string(buffer);
+		// std::cerr << "auth token: " << authentication_token << std::endl;
 		}
 	catch(Glib::SpawnError& err) {
 		std::cerr << "Failed to start server " << argv[0] << ": " << err.what() << std::endl;
@@ -216,8 +236,10 @@ void ComputeThread::on_open(websocketpp::connection_hdl )
 	{
 	connection_is_open=true;
 	restarting_kernel=false;
-	if(gui)
+	if(gui) {
 		gui->on_connect();
+		gui->on_kernel_runstatus(false);
+		}
 
 	//	// now it is safe to use the connection
 	//	std::string msg;
@@ -245,8 +267,13 @@ void ComputeThread::on_close(websocketpp::connection_hdl )
 	connection_is_open=false;
 	all_cells_nonrunning();
 	if(gui) {
-		if(restarting_kernel) gui->on_disconnect("restarting kernel");
-		else                  gui->on_disconnect("not connected");
+		if(restarting_kernel) {
+			gui->on_disconnect("restarting kernel");
+			gui->on_kernel_runstatus(true);
+			}
+		else {
+			gui->on_disconnect("not connected");
+			}
 		}
 
 	sleep(1); // do not cause a torrent...
@@ -267,46 +294,81 @@ void ComputeThread::cell_finished_running(DataCell::id_t id)
 void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 	{
 	client::connection_ptr con = wsclient.get_con_from_hdl(hdl);
-	//	std::cerr << msg->get_payload() << std::endl;
+	// std::cerr << "payload: " << msg->get_payload() << std::endl;
 
 	// Parse the JSON message.
-	Json::Value  root;
-	Json::Reader reader;
-	bool success = reader.parse( msg->get_payload(), root );
-	if ( !success ) {
-		std::cerr << "cadabra-client: cannot parse message" << std::endl;
+	nlohmann::json root;
+
+	try {
+		root=nlohmann::json::parse(msg->get_payload());
+		}
+	catch(nlohmann::json::exception& e) {
+		std::cerr << "cadabra-client: cannot parse message." << std::endl;
 		return;
 		}
-	const Json::Value header   = root["header"];
-	const Json::Value content  = root["content"];
-	const Json::Value msg_type = root["msg_type"];
+	if(root.count("header")==0) {
+		std::cerr << "cadabra-client: received message without 'header'." << std::endl;
+		return;
+		}
+	if(root.count("content")==0) {
+		std::cerr << "cadabra-client: received message without 'content'." << std::endl;
+		return;
+		}
+	const nlohmann::json& header  = root["header"];
+	const nlohmann::json& content = root["content"];
+	const std::string msg_type    = root.value("msg_type", "");
+
 	DataCell::id_t parent_id;
-	parent_id.id = header["parent_id"].asUInt64();
-	if(header["parent_origin"].asString()=="client")
+	parent_id.id = header.value("parent_id", uint64_t(0));
+	if(header.value("parent_origin", "")=="client")
 		parent_id.created_by_client=true;
 	else
 		parent_id.created_by_client=false;
 	DataCell::id_t cell_id;
-	cell_id.id = header["cell_id"].asUInt64();
-	if(header["cell_origin"].asString()=="client")
+	cell_id.id = header["cell_id"].get<uint64_t>();
+	if(header.value("cell_origin", "")=="client")
 		cell_id.created_by_client=true;
 	else
 		cell_id.created_by_client=false;
 	// std::cerr << "received cell with id " << cell_id.id << std::endl;
-	if (parent_id.id == interactive_cell) {
+
+	// Determine if this refers to a special cell in the interactive console.
+	if (interactive_cells.find(parent_id.id) != interactive_cells.end()) {
+		interactive_cells.insert(cell_id.id);
 		docthread->on_interactive_output(root);
-		console_child_ids.push_back(cell_id.id);
-		}
-	else if (cell_id.id == interactive_cell || std::find(console_child_ids.begin(), console_child_ids.end(), parent_id.id) != console_child_ids.end()) {
+	}
+	else if (interactive_cells.find(cell_id.id) != interactive_cells.end()) {
 		docthread->on_interactive_output(root);
 		}
-	else if (msg_type.asString().find("csl_") == 0) {
+	else if (msg_type.find("csl_") == 0) {
 		root["header"]["from_server"] = true;
 		docthread->on_interactive_output(root);
 		}
+	else if (msg_type == "progress_update") {
+		std::string msg = content.value<std::string>("msg", "Idle");
+		int n = content.value<int>("n", 0);
+		int total = content.value<int>("total", 0);
+		// FIXME: do something with 'pulse':
+		// int pulse = content.value<bool>("pulse", false);
+		docthread->set_progress(msg, n, total);
+	}
+	else if(msg_type=="completed") {
+		// std::cerr << "received completion of " << content["original"] << " -> " << content["completed"] << std::endl;
+
+		// Finally, the action to add the output cell.
+		std::string toadd=content["completed"].get<std::string>();
+		if(toadd.size()>0) {
+			toadd=toadd.substr(content["original"].get<std::string>().size());
+			int pos=content["position"].get<int>();
+			int alternative=content["alternative"].get<int>();
+			std::shared_ptr<ActionBase> action =
+				std::make_shared<ActionCompleteText>(cell_id, pos, toadd, alternative);
+			docthread->queue_action(action);
+			}
+		}
 	else {
 		try {
-			bool finished = header["last_in_sequence"].asBool();
+			bool finished = header["last_in_sequence"].get<bool>();
 
 			if (finished) {
 				std::shared_ptr<ActionBase> rs_action =
@@ -315,9 +377,9 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 				cell_finished_running(parent_id);
 				}
 
-			if (content["output"].asString().size() > 0) {
-				if (msg_type.asString() == "output") {
-					std::string output = "\\begin{verbatim}" + content["output"].asString() + "\\end{verbatim}";
+			if (content.count("output")>0 && content["output"].get<std::string>().size() > 0) {
+				if (msg_type == "output") {
+					std::string output = "\\begin{verbatim}" + content["output"].get<std::string>() + "\\end{verbatim}";
 
 					// Stick an AddCell action onto the stack. We instruct the
 					// action to add this result output cell as a child of the
@@ -329,8 +391,8 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
-				else if (msg_type.asString() == "verbatim") {
-					std::string output = "\\begin{verbatim}" + content["output"].asString() + "\\end{verbatim}";
+				else if (msg_type == "verbatim") {
+					std::string output = "\\begin{verbatim}" + content["output"].get<std::string>() + "\\end{verbatim}";
 
 					// Stick an AddCell action onto the stack. We instruct the
 					// action to add this result output cell as a child of the
@@ -342,23 +404,23 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
-				else if (msg_type.asString() == "latex_view") {
+				else if (msg_type == "latex_view") {
 					// std::cerr << "received latex cell " << content["output"].asString() << std::endl;
-					DataCell result(cell_id, DataCell::CellType::latex_view, content["output"].asString());
+					DataCell result(cell_id, DataCell::CellType::latex_view, content["output"].get<std::string>());
 					std::shared_ptr<ActionBase> action =
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
-				else if (msg_type.asString() == "input_form") {
-					DataCell result(cell_id, DataCell::CellType::input_form, content["output"].asString());
+				else if (msg_type == "input_form") {
+					DataCell result(cell_id, DataCell::CellType::input_form, content["output"].get<std::string>());
 					std::shared_ptr<ActionBase> action =
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
-				else if (msg_type.asString() == "error") {
-					std::string error = "{\\color{red}{\\begin{verbatim}" + content["output"].asString()
+				else if (msg_type == "error") {
+					std::string error = "{\\color{red}{\\begin{verbatim}" + content["output"].get<std::string>()
 					                    + "\\end{verbatim}}}";
-					if (msg_type.asString() == "fault") {
+					if (msg_type == "fault") {
 						error = "{\\color{red}{Kernel fault}}\\begin{small}" + error + "\\end{small}";
 						}
 
@@ -380,15 +442,21 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 
 					// FIXME: iterate over all cells and set the running flag to false.
 					}
-				else if (msg_type.asString() == "image_png") {
-					DataCell result(cell_id, DataCell::CellType::image_png, content["output"].asString());
+				else if (msg_type == "image_png") {
+					DataCell result(cell_id, DataCell::CellType::image_png, content["output"].get<std::string>());
+					std::shared_ptr<ActionBase> action =
+					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
+					docthread->queue_action(action);
+					}
+				else if (msg_type == "image_svg") {
+					DataCell result(cell_id, DataCell::CellType::image_svg, content["output"].get<std::string>());
 					std::shared_ptr<ActionBase> action =
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
 				else {
 					std::cerr << "cadabra-client: received cell we did not expect: "
-					          << msg_type.asString() << std::endl;
+					          << msg_type << std::endl;
 					}
 				}
 			}
@@ -407,34 +475,31 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 	gui->process_data();
 	}
 
-void ComputeThread::execute_interactive(const std::string& code)
+void ComputeThread::execute_interactive(uint64_t id, const std::string& code)
 	{
 	assert(gui_thread_id == std::this_thread::get_id());
 
-	if (!connection_is_open || interactive_cell == 0)
+	if (!connection_is_open)
 		return;
 
 	if (code.substr(0, 7) == "reset()")
 		return restart_kernel();
 
-	Json::Value req, header, content;
+	nlohmann::json req, header, content;
 
-	header["msg_type"] = "execute_request";
-	header["cell_id"] = static_cast<Json::UInt64>(interactive_cell);
+	header["msg_type"]    = "execute_request";
+	header["cell_id"]     = id;
 	header["interactive"] = true;
-	content["code"] = code.c_str();
+	content["code"]       = code.c_str();
 
-	req["header"] = header;
-	req["content"] = content;
+	req["auth_token"] = authentication_token;
+	req["header"]     = header;
+	req["content"]    = content;
 
 	std::ostringstream oss;
 	oss << req << std::endl;
 	wsclient.send(our_connection_hdl, oss.str(), websocketpp::frame::opcode::text);
-	}
-
-void ComputeThread::register_interactive_cell(uint64_t id)
-	{
-	interactive_cell = id;
+	interactive_cells.insert(id);
 	}
 
 void ComputeThread::execute_cell(DTree::iterator it)
@@ -476,14 +541,15 @@ void ComputeThread::execute_cell(DTree::iterator it)
 		   std::make_shared<ActionSetRunStatus>(it->id(), true);
 		docthread->queue_action(rs_action);
 
-		Json::Value req, header, content;
+		nlohmann::json req, header, content;
 		header["uuid"]="none";
-		header["cell_id"]=(Json::UInt64)dc.id().id;
+		header["cell_id"]=dc.id().id;
 		if(dc.id().created_by_client)
 			header["cell_origin"]="client";
 		else
 			header["cell_origin"]="server";
 		header["msg_type"]="execute_request";
+		req["auth_token"]=authentication_token;
 		req["header"]=header;
 		content["code"]=dc.textbuf;
 		req["content"]=content;
@@ -519,9 +585,10 @@ void ComputeThread::stop()
 	if(connection_is_open==false)
 		return;
 
-	Json::Value req, header, content;
+	nlohmann::json req, header, content;
 	header["uuid"]="none";
 	header["msg_type"]="execute_interrupt";
+	req["auth_token"]=authentication_token;
 	req["header"]=header;
 
 	std::ostringstream str;
@@ -531,6 +598,7 @@ void ComputeThread::stop()
 
 	server_pid=0;
 	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+	all_cells_nonrunning();	
 	}
 
 void ComputeThread::restart_kernel()
@@ -546,10 +614,11 @@ void ComputeThread::restart_kernel()
 	gui->on_kernel_runstatus(false);
 
 	// std::cerr << "cadabra-client: restarting kernel" << std::endl;
-	Json::Value req, header, content;
+	nlohmann::json req, header, content;
 	header["uuid"]="none";
 	header["msg_type"]="exit";
 	header["from_server"] = true;
+	req["auth_token"]=authentication_token;
 	req["header"]=header;
 
 	std::ostringstream str;
@@ -560,3 +629,47 @@ void ComputeThread::restart_kernel()
 	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
 	docthread->on_interactive_output(req);
 	}
+
+bool ComputeThread::complete(DTree::iterator it, int pos, int alternative)
+	{
+	if(connection_is_open==false)
+		return false;
+
+	const DataCell& dc=(*it);
+
+	nlohmann::json req, header, content;
+	header["uuid"]="none";
+	header["cell_id"]=dc.id().id;
+	if(dc.id().created_by_client)
+		header["cell_origin"]="client";
+	else
+		header["cell_origin"]="server";
+	header["msg_type"]="complete";
+	req["auth_token"]=authentication_token;
+	req["header"]=header;
+	std::string todo = it->textbuf.substr(0,pos);
+//	if(todo.size()>0 && todo[todo.size()-1]=='\n')
+//		todo=todo.substr(0, todo.size()-1);
+	// std::cerr << "to complete full: " << todo << std::endl;
+	size_t lst=todo.find_last_of("\n(){}[]:\t ");
+	if(lst!=std::string::npos)
+		todo=todo.substr(lst+1);
+	// std::cerr << "to complete strip: " << todo << std::endl;
+
+	if(todo.size()==0)
+		return false;
+
+	req["string"]=todo;
+	req["position"]=pos;
+	req["alternative"]=alternative;
+
+	std::ostringstream str;
+	str << req << std::endl;
+
+	// std::cerr << str.str() << std::endl;
+
+	server_pid=0;
+	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+
+	return true;
+	}
