@@ -2,17 +2,32 @@
 #include "InstallPrefix.hh"
 #include "Config.hh"
 #include "DataCell.hh"
+#include "Exceptions.hh"
 #include <sstream>
 #include <fstream>
 #include <regex>
-#include <boost/filesystem.hpp>
-#include <boost/algorithm/string.hpp>
+//#include <boost/filesystem.hpp>
+//#include <boost/algorithm/string.hpp>
 #include <iostream>
-#include <glibmm/base64.h>
+#include "base64.hh"
 #include <internal/uuid.h>
+#include <iomanip>
+
 
 using namespace cadabra;
 
+// General tool to strip spaces from both ends
+static std::string trim(const std::string& s)
+	{
+	if(s.length() == 0)
+		return s;
+	int b = s.find_first_not_of(" \t\n");
+	int e = s.find_last_not_of(" \t\n");
+	if(b == -1) // No non-spaces
+		return "";
+	return std::string(s, b, e - b + 1);
+	}
+
 bool DataCell::id_t::operator<(const DataCell::id_t& other) const
 	{
 	if(created_by_client != other.created_by_client) return created_by_client;
@@ -21,7 +36,7 @@ bool DataCell::id_t::operator<(const DataCell::id_t& other) const
 	}
 
 DataCell::id_t::id_t()
-	: id(generate_uuid<Json::UInt64>())
+	: id(generate_uuid<uint64_t>())
 	, created_by_client(true)
 	{
 	}
@@ -32,6 +47,7 @@ DataCell::DataCell(CellType t, const std::string& str, bool cell_hidden)
 	textbuf = str;
 	hidden = cell_hidden;
 	running=false;
+	ignore_on_import=false;
 	}
 
 DataCell::DataCell(id_t id_, CellType t, const std::string& str, bool cell_hidden)
@@ -41,6 +57,7 @@ DataCell::DataCell(id_t id_, CellType t, const std::string& str, bool cell_hidde
 	hidden = cell_hidden;
 	running=false;
 	serial_number=id_;
+	ignore_on_import=false;	
 	}
 
 DataCell::DataCell(const DataCell& other)
@@ -51,6 +68,7 @@ DataCell::DataCell(const DataCell& other)
 	hidden = other.hidden;
 	sensitive = other.sensitive;
 	serial_number = other.serial_number;
+	ignore_on_import = other.ignore_on_import;
 	}
 
 std::string cadabra::export_as_HTML(const DTree& doc, bool for_embedding, bool strip_code, std::string title)
@@ -79,6 +97,7 @@ std::string cadabra::latex_to_html(const std::string& str)
 	std::regex discretionary(R"(\\discretionary\{\}\{\}\{\})");
 	std::regex subsection(R"(\\subsection\*\{([^\}]*)\})");
 	std::regex subsubsection(R"(\\subsubsection\*\{([^\}]*)\})");
+	std::regex emph(R"(\\emph\{([^\}]*)\})");	
 	std::regex verb(R"(\\verb\|([^\|]*)\|)");
 	std::regex url(R"(\\url\{([^\}]*)\})");
 	std::regex href(R"(\\href\{([^\}]*)\}\{([^\}]*)\})");
@@ -89,6 +108,7 @@ std::string cadabra::latex_to_html(const std::string& str)
 	std::regex tilde("~");
 	std::regex less("<");
 	std::regex greater(">");
+	std::regex bigO(R"(\\bigO)");
 	std::regex latex(R"(\\LaTeX\{\})");
 	std::regex tex(R"(\\TeX\{\})");
 	std::regex algorithm(R"(\\algorithm\{([^\}]*)\}\{([^\}]*)\})");
@@ -106,20 +126,24 @@ std::string cadabra::latex_to_html(const std::string& str)
 	std::regex ftableau(R"(\\ftableau\{(\{[^\}]*\}[,]?)*\})");
 	std::regex begin_tabular(R"(\\begin\{tabular\}\{[^\}]*\})");
 	std::regex end_tabular(R"(\\end\{tabular\})");
+	std::regex hash_in_math(R"((\\begin\{dmath\*\}.*)\\#(\\end\{dmath\*\}))");
 	std::string res;
 
 	try {
+		res = std::regex_replace(res, hash_in_math, "$1#$2");
 		res = std::regex_replace(str, begin_dmath, R"(\(\displaystyle)");
 		res = std::regex_replace(res, discretionary, " ");
 		res = std::regex_replace(res, end_dmath, R"(\))");
 		res = std::regex_replace(res, tilde, " ");
 		res = std::regex_replace(res, less, "&lt;");
-		res = std::regex_replace(res, tilde, "&gt;");
+		res = std::regex_replace(res, greater, "&gt;");
+		res = std::regex_replace(res, bigO, "{\\cal O}");		
 		res = std::regex_replace(res, begin_verbatim, "<pre class='output'>");
 		res = std::regex_replace(res, end_verbatim, "</pre>");
 		res = std::regex_replace(res, section, "<h1>$1</h1>");
 		res = std::regex_replace(res, subsection, "<h2>$1</h2>");
 		res = std::regex_replace(res, subsubsection, "<h3>$1</h3>");
+		res = std::regex_replace(res, emph, "<i>$1</i>");
 		res = std::regex_replace(res, author, "<div class='author'>$1</div>");
 		res = std::regex_replace(res, email, "<div class='email'>$1</div>");
 		res = std::regex_replace(res, verb, "<code>$1</code>");
@@ -194,6 +218,9 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 		case DataCell::CellType::image_png:
 			str << "<div class='image_png'><img src='data:image/png;base64,";
 			break;
+		case DataCell::CellType::image_svg:
+			str << "<div class='image_svg'><img src='data:image/svg+xml;base64,";
+			break;
 		case DataCell::CellType::input_form:
 			str << "<div class='input_form'>";
 			break;
@@ -204,14 +231,21 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 			if(it->textbuf.size()>0) {
 				if(it->cell_type==DataCell::CellType::image_png)
 					str << it->textbuf;
+				else if(it->cell_type==DataCell::CellType::image_svg)
+					str << it->textbuf;
 				else if(it->cell_type!=DataCell::CellType::document && it->cell_type!=DataCell::CellType::latex) {
 					std::string out;
 					if(it->cell_type==DataCell::CellType::python)
 						out=it->textbuf;
 					else
 						out=latex_to_html(it->textbuf);
-					if(out.size()>0)
+					if(out.size()>0) {
+						if(it->cell_type==DataCell::CellType::python) {
+							out = std::regex_replace(out, std::regex("<"), "&lt;");
+							out = std::regex_replace(out, std::regex(">"), "&gt;");
+							}
 						str << "<div class=\"source donthyphenate\">"+out+"</div>";
+						}
 					}
 				}
 			}
@@ -262,6 +296,9 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 		case DataCell::CellType::image_png:
 			str << "' /></div>\n";
 			break;
+		case DataCell::CellType::image_svg:
+			str << "' /></div>\n";
+			break;
 		case DataCell::CellType::input_form:
 			str << "</div>\n";
 		}
@@ -269,16 +306,16 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 
 std::string cadabra::JSON_serialise(const DTree& doc)
 	{
-	Json::Value json;
+	nlohmann::json json;
 	JSON_recurse(doc, doc.begin(), json);
 
 	std::ostringstream str;
-	str << json;
+	str << std::setfill('\t') << std::setw(1) << json;
 
 	return str.str();
 	}
 
-void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& json)
+void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, nlohmann::json& json)
 	{
 	switch(it->cell_type) {
 		case DataCell::CellType::document:
@@ -306,6 +343,9 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& js
 		case DataCell::CellType::image_png:
 			json["cell_type"]="image_png";
 			break;
+		case DataCell::CellType::image_svg:
+			json["cell_type"]="image_svg";
+			break;
 		case DataCell::CellType::input_form:
 			json["cell_type"]="input_form";
 			break;
@@ -313,7 +353,7 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& js
 			//			assert(1==0);
 			//			// NOT YET FUNCTIONAL
 			//			json["cell_type"]="section";
-			//			Json::Value child;
+			//			nlohmann::json child;
 			//			child["content"]="test";
 			//			json.append(child);
 			//			break;
@@ -321,6 +361,8 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& js
 		}
 	if(it->hidden)
 		json["hidden"]=true;
+	if(it->ignore_on_import)
+		json["ignore_on_import"]=true;
 
 	json["cell_id"] = it->id().id;
 
@@ -334,12 +376,12 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& js
 		}
 
 	if(doc.number_of_children(it)>0) {
-		Json::Value cells;
+		nlohmann::json cells=nlohmann::json::array();
 		DTree::sibling_iterator sib=doc.begin(it);
 		while(sib!=doc.end(it)) {
-			Json::Value thiscell;
+			nlohmann::json thiscell;
 			JSON_recurse(doc, sib, thiscell);
-			cells.append(thiscell);
+			cells.push_back(thiscell);
 			++sib;
 			}
 		json["cells"]=cells;
@@ -348,120 +390,139 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, Json::Value& js
 
 void cadabra::JSON_deserialise(const std::string& cj, DTree& doc)
 	{
-	Json::Reader reader;
-	Json::Value  root;
+	nlohmann::json  root;
 
-	bool ret = reader.parse( cj, root );
-	if ( !ret ) {
-		std::cerr << "cannot parse json file" << std::endl;
+	try {
+		root=nlohmann::json::parse(cj);
+		}
+	catch(nlohmann::json::exception& e) {
+		std::cerr << "Cannot parse json file." << std::endl;
 		return;
 		}
 
 	// Setup main document.
-	DataCell top(DataCell::CellType::document);
+	DataCell::id_t id;
+	try {
+		id.id=root.value("cell_id", generate_uuid<uint64_t>());
+		}
+	catch(nlohmann::json::exception& e) {
+		std::cerr << "Failed to find root-level 'cell_id' element (must be an integer)." << std::endl;
+		return;
+		}
+	DataCell top(id, DataCell::CellType::document);
 	DTree::iterator doc_it = doc.set_head(top);
 
+	// Determine whether this is a Jupyter notebook or a Cadabra
+	// notebook.
+	if(root.count("description") == 0) {
+		root = cadabra::ipynb2cnb(root);
+		}
+	
 	// Scan through json file.
-	const Json::Value cells = root["cells"];
+	const nlohmann::json& cells = root["cells"];
 	JSON_in_recurse(doc, doc_it, cells);
 	}
 
-void cadabra::JSON_in_recurse(DTree& doc, DTree::iterator loc, const Json::Value& cells)
+void cadabra::JSON_in_recurse(DTree& doc, DTree::iterator loc, const nlohmann::json& cells)
 	{
 	try {
-		for(unsigned int c=0; c<cells.size(); ++c) {
-			const Json::Value celltype    = cells[c]["cell_type"];
-			const Json::Value cell_id     = cells[c].get("cell_id", generate_uuid<Json::UInt64>()).asUInt64();
-			const Json::Value cell_origin = cells[c]["cell_origin"];
-			const Json::Value textbuf     = cells[c]["source"];
-			const Json::Value hidden      = cells[c]["hidden"];
+		for(const auto& cell: cells) {
+			const nlohmann::json& textbuf = cell["source"];
 
 			DTree::iterator last=doc.end();
 			DataCell::id_t id;
-			id.id=cell_id.asUInt64();
-			if(cell_origin=="server")
+			id.id=cell.value("cell_id", generate_uuid<uint64_t>());
+
+			if(cell.value("cell_origin", "")=="server")
 				id.created_by_client=false;
 			else
 				id.created_by_client=true;
-
+			
 			bool hide=false;
-			if(hidden.asBool())
+			if(cell.value("hidden", false))
 				hide=true;
 
-			if(celltype.asString()=="input" || celltype.asString()=="code") {
+			std::string cell_type = cell.value("cell_type", "");
+			if(cell_type=="input" || cell_type=="code") {
 				std::string res;
-				if(textbuf.isArray()) {
-					for(auto& el: textbuf)
-						res+=el.asString();
+				if(textbuf.is_array()) {
+					for(const auto& el: textbuf)
+						res+=el.get<std::string>();
 					}
 				else {
-					res=textbuf.asString();
+					res=textbuf.get<std::string>();
 					}
 				DataCell dc(id, cadabra::DataCell::CellType::python, res, hide);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="output") {
-				DataCell dc(id, cadabra::DataCell::CellType::output, textbuf.asString(), hide);
+			else if(cell_type=="output") {
+				DataCell dc(id, cadabra::DataCell::CellType::output, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="error") {
-				DataCell dc(id, cadabra::DataCell::CellType::error, textbuf.asString(), hide);
+			else if(cell_type=="error") {
+				DataCell dc(id, cadabra::DataCell::CellType::error, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="verbatim") {
-				DataCell dc(id, cadabra::DataCell::CellType::verbatim, textbuf.asString(), hide);
+			else if(cell_type=="verbatim") {
+				DataCell dc(id, cadabra::DataCell::CellType::verbatim, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="input_form") {
-				DataCell dc(id, cadabra::DataCell::CellType::input_form, textbuf.asString(), hide);
+			else if(cell_type=="input_form") {
+				DataCell dc(id, cadabra::DataCell::CellType::input_form, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="latex_view") {
+			else if(cell_type=="latex_view") {
 				std::string res;
-				if(textbuf.isArray()) {
+				if(textbuf.is_array()) {
 					for(auto& el: textbuf)
-						res+=el.asString();
+						res+=el.get<std::string>();
 					}
 				else {
-					res=textbuf.asString();
+					res=textbuf.get<std::string>();
 					}
 				DataCell dc(id, cadabra::DataCell::CellType::latex_view, res, false);
 				last=doc.append_child(loc, dc);
 				}
-			else if(celltype.asString()=="latex" || celltype.asString()=="markdown") {
+			else if(cell_type=="latex" || cell_type=="markdown") {
 				std::string res;
-				if(textbuf.isArray()) {
+				if(textbuf.is_array()) {
 					for(auto& el: textbuf)
-						res+=el.asString();
+						res+=el.get<std::string>();
 					}
 				else {
-					res=textbuf.asString();
+					res=textbuf.get<std::string>();
 					}
 				bool hide_jupyter=hide;
-				if(cells[c].isMember("cells")==false) hide_jupyter=true;
+				if(cell.count("cells")==0) hide_jupyter=true;
 
 				DataCell dc(id, cadabra::DataCell::CellType::latex, res, hide_jupyter);
 				last=doc.append_child(loc, dc);
 
 				// IPython/Jupyter notebooks only have the input LaTeX cell, not the output cell,
 				// which we need.
-				if(cells[c].isMember("cells")==false) {
+				if(cell.count("cells")==0) {
 					DataCell dc(id, cadabra::DataCell::CellType::latex_view, res, hide);
 					doc.append_child(last, dc);
 					}
 				}
-			else if(celltype.asString()=="image_png") {
-				DataCell dc(id, cadabra::DataCell::CellType::image_png, textbuf.asString(), hide);
+			else if(cell_type=="image_png") {
+				DataCell dc(id, cadabra::DataCell::CellType::image_png, textbuf.get<std::string>(), hide);
+				last=doc.append_child(loc, dc);
+				}
+			else if(cell_type=="image_svg") {
+				DataCell dc(id, cadabra::DataCell::CellType::image_svg, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
 			else {
-				std::cerr << "cadabra-client: found unknown cell type '"+celltype.asString()+"', ignoring" << std::endl;
+				std::cerr << "cadabra-client: found unknown cell type '"+cell_type+"', ignoring" << std::endl;
 				continue;
 				}
 
 			if(last!=doc.end()) {
-				if(cells[c].isMember("cells")) {
-					const Json::Value subcells = cells[c]["cells"];
+				if(cell.value("ignore_on_import", false))
+					last->ignore_on_import=true;
+				if(cell.count("cells")>0) {
+					const nlohmann::json& subcells = cell["cells"];
 					JSON_in_recurse(doc, last, subcells);
 					}
 				}
@@ -477,34 +538,39 @@ DataCell::id_t DataCell::id() const
 	return serial_number;
 	}
 
-std::string cadabra::export_as_LaTeX(const DTree& doc, const std::string& image_file_base)
+std::string cadabra::export_as_LaTeX(const DTree& doc, const std::string& image_file_base, bool for_embedding)
 	{
 	// Load the pre-amble from file.
-	std::string pname = cadabra::install_prefix()+"/share/cadabra2/notebook.tex";
-	std::ifstream preamble(pname);
-	if(!preamble)
-		throw std::logic_error("Cannot open LaTeX preamble at "+pname);
-	std::stringstream buffer;
-	buffer << preamble.rdbuf();
-	// std::cerr << "Using preamble at " << pname << std::endl;
-	std::string preamble_string = buffer.str();
+	std::string preamble_string;
+	if(!for_embedding) {
+		std::string pname = cadabra::install_prefix()+"/share/cadabra2/notebook.tex";
+		std::ifstream preamble(pname);
+		if(!preamble)
+			throw std::logic_error("Cannot open LaTeX preamble at "+pname);
+		std::stringstream buffer;
+		buffer << preamble.rdbuf();
+		// std::cerr << "Using preamble at " << pname << std::endl;
+		preamble_string = buffer.str();
+		}
 
 	// Open the LaTeX file for writing.
 	std::ostringstream str;
 	int image_num=0;
-	LaTeX_recurse(doc, doc.begin(), str, preamble_string, image_file_base, image_num);
+	LaTeX_recurse(doc, doc.begin(), str, preamble_string, image_file_base, image_num, for_embedding);
 
 	return str.str();
 	}
 
 void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringstream& str,
                             const std::string& preamble_string, const std::string& image_file_base,
-                            int& image_num)
+                            int& image_num, bool for_embedding)
 	{
 	switch(it->cell_type) {
 		case DataCell::CellType::document:
-			str << preamble_string;
-			str << "\\begin{document}\n";
+			if(!for_embedding) {
+				str << preamble_string;
+				str << "\\begin{document}\n";
+				}
 			break;
 		case DataCell::CellType::python:
 			str << "\\begin{python}\n";
@@ -513,7 +579,7 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 			str << "\\begin{python}\n";
 			break;
 		case DataCell::CellType::verbatim:
-			str << "\\begin{verbatim}\n";
+//			str << "\\begin{verbatim}\n";
 			break;
 		case DataCell::CellType::latex:
 			break;
@@ -523,7 +589,8 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 			break;
 		case DataCell::CellType::input_form:
 			break;
-		case DataCell::CellType::image_png:
+ 		case DataCell::CellType::image_png:
+ 		case DataCell::CellType::image_svg:
 			std::size_t pos=image_file_base.rfind('/');
 			std::string fileonly=image_file_base.substr(pos+1);
 			str << "\\begin{center}\n\\includegraphics[width=.6\\textwidth]{"
@@ -536,7 +603,21 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		// Images have to be saved to disk as separate files as
 		// LaTeX has no concept of images embedded in the .tex file.
 		std::ofstream out(image_file_base+std::to_string(image_num)+".png");
-		out << Glib::Base64::decode(it->textbuf);
+		out << base64_decode(it->textbuf);
+		++image_num;
+		}
+	else if(it->cell_type==DataCell::CellType::image_svg) {
+		// Images have to be saved to disk as separate files as
+		// LaTeX has no concept of images embedded in the .tex file.
+		std::string basename=image_file_base+std::to_string(image_num);
+		std::ofstream out(basename+".svg");
+		out << base64_decode(it->textbuf);
+		out.close();
+		// Convert to pdf, otherwise LaTeX cannot include it.
+      //		inkscape t.svg --export-pdf=t.pdf
+		int res = system((std::string("inkscape ")+basename+std::string(".svg --export-pdf=")+basename+std::string(".pdf")).c_str());
+		if(res!=0)
+			throw std::logic_error("DataCell::LaTeX_recurse: failed to run 'convert' to convert svg to pdf.");
 		++image_num;
 		}
 	else {
@@ -554,10 +635,13 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 				lr=std::regex_replace(lr, std::regex(R"(\\left\\\{)"),            "\\brwrap{\\{}{");
 				lr=std::regex_replace(lr, std::regex(R"(\\right\\\})"),           "}{\\}}");
 				lr=std::regex_replace(lr, std::regex(R"(\\right\.)"),            "}{.}");
-				lr=std::regex_replace(lr, std::regex(R"(\\begin\{verbatim\})"), "");
-				lr=std::regex_replace(lr, std::regex(R"(\\end\{verbatim\})"),   "");
+//				lr=std::regex_replace(lr, std::regex(R"(\\begin\{verbatim\})"), "");
+//				lr=std::regex_replace(lr, std::regex(R"(\\end\{verbatim\})"),   "");
 				lr=std::regex_replace(lr, std::regex(R"(\\begin\{dmath\*\})"),  "\\begin{adjustwidth}{1em}{0cm}$");
 				lr=std::regex_replace(lr, std::regex(R"(\\end\{dmath\*\})"),    "$\\end{adjustwidth}");
+				lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");
+				lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");
+				lr=std::regex_replace(lr, std::regex(R"(🍅)"), "*");
 				str << lr << "\n";
 				}
 			}
@@ -569,7 +653,7 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 			str << "\\end{python}\n";
 			break;
 		case DataCell::CellType::verbatim:
-			str << "\\end{verbatim}\n";
+//			str << "\\end{verbatim}\n";
 			break;
 		case DataCell::CellType::document:
 		case DataCell::CellType::latex:
@@ -577,20 +661,23 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		case DataCell::CellType::input_form:
 		case DataCell::CellType::error:
 		case DataCell::CellType::image_png:
+		case DataCell::CellType::image_svg:
 			break;
 		}
 
 	if(doc.number_of_children(it)>0) {
 		DTree::sibling_iterator sib=doc.begin(it);
 		while(sib!=doc.end(it)) {
-			LaTeX_recurse(doc, sib, str, preamble_string, image_file_base, image_num);
+			LaTeX_recurse(doc, sib, str, preamble_string, image_file_base, image_num, for_embedding);
 			++sib;
 			}
 		}
 
 	switch(it->cell_type) {
 		case DataCell::CellType::document:
-			str << "\\end{document}\n";
+			if(!for_embedding) {
+				str << "\\end{document}\n";
+				}
 			break;
 		case DataCell::CellType::python:
 		case DataCell::CellType::output:
@@ -600,6 +687,7 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		case DataCell::CellType::input_form:
 		case DataCell::CellType::error:
 		case DataCell::CellType::image_png:
+		case DataCell::CellType::image_svg:
 			break;
 		}
 
@@ -645,3 +733,125 @@ void cadabra::python_recurse(const DTree& doc, DTree::iterator it, std::ostrings
 // 	   }
 //    return str;
 //    }
+
+nlohmann::json cadabra::ipynb2cnb(const nlohmann::json& root)
+	{
+	int nbf = root.value("nbformat", 0);
+	nlohmann::json json;
+
+	if(nbf==0)
+		throw RuntimeException("Not a Jupyter notebook.");
+
+	json["description"]="Cadabra JSON notebook format";
+	json["version"]=1.0;
+
+	nlohmann::json cells=nlohmann::json::array();
+	const nlohmann::json& jucells=root["cells"];
+	
+	// Jupyter notebooks just have a single array of cells; walk
+	// through and add to our "cells" array.
+
+	for(unsigned int c=0; c<jucells.size(); ++c) {
+		nlohmann::json cell;
+		if(jucells[c]["cell_type"].get<std::string>()=="markdown")
+			cell["cell_type"]="latex";
+		else
+			cell["cell_type"]="input";
+		cell["hidden"]=false;
+		const nlohmann::json& source=jucells[c]["source"];
+		// Jupyter stores the source line-by-line in an array 'source'.
+		std::string block;
+		for(unsigned int l=0; l<source.size(); ++l) {
+			std::string line=source[l].get<std::string>();
+			if(line.size()>0) {
+				if(line[0]=='#') {
+					std::string sub="";
+					line=line.substr(1);
+					int inc=0;
+					while(line.size()>0 && line[0]=='#') {
+						if(inc<2) {
+							sub+="sub";
+							inc+=1;
+							}
+						line=line.substr(1);
+						}
+					if(line.size()==0)
+						continue;
+					if(line[line.size()-1]=='\n')
+						line=line.substr(0,line.size()-1);
+					block+="\\"+sub+"section*{"+trim(line)+"}\n";
+					}
+				else
+					block+=line;
+				}
+			}
+		cell["source"]=block;
+		cells.push_back(cell);
+		}
+
+	json["cells"] = cells;
+	
+	return json;
+	}
+
+nlohmann::json cadabra::cnb2ipynb(const nlohmann::json& root)
+	{
+	nlohmann::json ipynb, kernelspec, lang;
+
+	ipynb["nbformat"]=4;
+	ipynb["nbformat_minor"]=4;
+
+	kernelspec["display_name"]="Cadabra2";
+	kernelspec["language"]="python";
+	kernelspec["name"]="cadabra2";
+
+	lang["codemirror_mode"]="cadabra";
+	lang["file_extension"]=".ipynb";
+	lang["mimetype"]="text/cadabra";
+	lang["name"]="cadabra2";
+	lang["pygments_lexer"]="cadabra";
+	
+	ipynb["metadata"] = {
+		{"kernelspec", kernelspec},
+		{"language_info", lang}
+	};
+
+	nlohmann::json cells=nlohmann::json::array();
+	
+	// Jupyter notebooks just have a single array of cells; walk
+	// through and setup our cells array.
+
+	for(const auto& cdb: root["cells"]) {
+		nlohmann::json cell;
+		if(cdb["cell_type"]=="python") {
+			cell["cell_type"]="code";
+			cell["source"]=nlohmann::json::array();
+			cell["source"].push_back(cdb["source"]);
+			cell["metadata"]=nlohmann::json::object();
+			cell["outputs"]=nlohmann::json::array();
+			cell["execution_count"]={};			
+			}
+		if(cdb["cell_type"]=="input") {
+			cell["cell_type"]="code";
+			cell["source"]=nlohmann::json::array();
+			cell["source"].push_back(cdb["source"]);
+			cell["metadata"]=nlohmann::json::object();
+			cell["outputs"]=nlohmann::json::array();
+			cell["execution_count"]={};
+			}
+		else if(cdb["cell_type"]=="latex") {
+			cell["cell_type"]="markdown";
+			cell["source"]=nlohmann::json::array();
+			cell["source"].push_back(cdb["source"]);
+			cell["metadata"]=nlohmann::json::object();
+			}
+
+		if(cell.is_null()==false)
+			cells.push_back(cell);
+		
+		}
+
+	ipynb["cells"] = cells;
+	
+	return ipynb;
+	}
