@@ -23,8 +23,12 @@ You should have received a copy of the GNU General Public License
 #include "Exceptions.hh"
 #include <iomanip>
 #include <sstream>
-#include <pcrecpp.h>
+#include <locale>
+#include <codecvt>
 
+#include <boost/functional/hash.hpp>
+
+//#define DEBUG 1
 
 namespace cadabra {
 
@@ -36,6 +40,11 @@ namespace cadabra {
 		return mul.get_num().get_si();
 		}
 
+	double to_double(multiplier_t mul)
+		{
+		return mul.get_d();
+		}
+
 	std::string to_string(long num)
 		{
 		std::ostringstream str;
@@ -61,11 +70,22 @@ namespace cadabra {
 		}
 
 	Ex::Ex(const Ex& other)
-		: tree<str_node>(other), state_(result_t::l_no_action)
+		: std::enable_shared_from_this<Ex>(other), tree<str_node>(other), state_(result_t::l_no_action)
 		{
 		//	std::cout << "Ex copy constructor" << std::endl;
 		}
 
+	Ex& Ex::operator=(Ex other)
+		{
+		std::swap(static_cast<tree<str_node>&>(*this),
+					 static_cast<tree<str_node>&>(other));
+		std::swap((*this).state_, other.state_);
+		std::swap((*this).history, other.history);
+		std::swap((*this).terms, other.terms);
+
+		return *this;
+		}
+
 	Ex::Ex(const std::string& str)
 		: state_(result_t::l_no_action)
 		{
@@ -79,6 +99,14 @@ namespace cadabra {
 		multiply(begin()->multiplier, val);
 		}
 
+	Ex::Ex(float val)
+		: state_(result_t::l_no_action)
+		{
+		std::ostringstream str;
+		str << val;
+		set_head(str_node(str.str()));
+		}
+
 	Ex::result_t Ex::state() const
 		{
 		return state_;
@@ -127,6 +155,26 @@ namespace cadabra {
 		return *(begin()->multiplier);
 		}
 
+	bool Ex::is_integer() const
+		{
+		if(begin()!=end())
+			if(begin()->is_integer())
+				return true;
+		return false;
+		}
+
+	long Ex::to_integer() const
+		{
+		if(!is_integer())
+			throw InternalError("Called to_integer() on non-integer Ex");
+		return to_long(*(begin()->multiplier));
+		}
+
+	bool Ex::is_empty() const
+		{
+		return begin()==end();
+		}
+
 	std::ostream& Ex::print_python(std::ostream& str, Ex::iterator it)
 		{
 		std::string name(*(*it).name);
@@ -305,7 +353,7 @@ namespace cadabra {
 		//	str << "  (" << calc_hash(it) << ")";
 		//	str << "  (" << depth(it) << ")";
 		//	str << "  (" << it->fl.bracket << " " << &(*it) << ")";
-		str << "  (" << it->fl.bracket << " " << it.node << ")";
+//		str << "  (" << it->fl.bracket << " " << it.node << ")";
 		if(!compact_tree) str << std::endl;
 
 		while(beg!=fin) {
@@ -382,18 +430,20 @@ namespace cadabra {
 		// the correct thing to do?)
 		//
 		// If this algorithm is changed, factorise::calc_restricted_hash in
-		// modules/algebra.cc should also be modified!
+		// core/algorithms/factor_in.cc should also be modified!
 
-		hashval_t ret=(hashval_t)(&(*it->name));
+		iterator end=it;
+		end.skip_children();
+		++end;
+		it.skip_children(false);
 
-		sibling_iterator sub=begin(it);
-		while(sub!=end(it)) {
-			ret*=17;
-			ret+=calc_hash(sub);
-			++sub;
+		hashval_t seed = 0;
+		while(it!=end) {
+			boost::hash_combine(seed, *it->name);
+			++it;
 			}
 
-		return ret;
+		return seed;
 		}
 
 	Ex::sibling_iterator Ex::arg(iterator it, unsigned int num)
@@ -670,6 +720,12 @@ found:
 		return equal_subtree(begin(), other.begin());
 		}
 
+//	bool Ex::operator<(const Ex& other) const
+//		{
+//		Ex::iterator_base_less less;
+//		return less(begin(), other.begin());
+//		}
+
 	void Ex::push_history(const std::vector<Ex::path_t>& paths)
 		{
 		history.push_back(*this);
@@ -709,6 +765,27 @@ found:
 		//	fl.mark=0;
 		}
 
+	str_node::str_node(const std::u32string& nm, bracket_t br, parent_rel_t pr)
+		{
+#ifdef _MSC_VER
+		std::wstring_convert<std::codecvt_utf8<int32_t>, int32_t> conv;
+		auto p = reinterpret_cast<const int32_t*>(nm.data());
+		std::string nm8 = conv.to_bytes(p, p + nm.size());
+#else
+		std::wstring_convert<std::codecvt_utf8<char32_t>, char32_t> conv;
+		std::string nm8=conv.to_bytes(nm);
+#endif
+
+#ifdef DEBUG
+		std::cerr << "str_node: " << nm8 << std::endl;
+#endif
+		multiplier=rat_set.insert(1).first;
+		name=name_set.insert(nm8).first;
+		//	fl.modifier=m_none;
+		fl.bracket=br;
+		fl.parent_rel=pr;
+		}
+
 	str_node::str_node(const std::string& nm, bracket_t br, parent_rel_t pr)
 		{
 		multiplier=rat_set.insert(1).first;
@@ -952,12 +1029,20 @@ found:
 	void multiply(rset_t::iterator& num, multiplier_t fac)
 		{
 		fac*=*num;
+		fac.canonicalize();
+		num=rat_set.insert(fac).first;
+		}
+
+	void set(rset_t::iterator& num, multiplier_t fac)
+		{
+		fac.canonicalize();
 		num=rat_set.insert(fac).first;
 		}
 
 	void add(rset_t::iterator& num, multiplier_t fac)
 		{
 		fac+=*num;
+		fac.canonicalize();
 		num=rat_set.insert(fac).first;
 		}
 
@@ -973,18 +1058,23 @@ found:
 
 	void flip_sign(rset_t::iterator& num)
 		{
-		num=rat_set.insert(-(*num)).first;
+		multiplier_t fac=-(*num);
+		fac.canonicalize();
+		num=rat_set.insert(fac).first;
 		}
 
 	void half(rset_t::iterator& num)
 		{
-		num=rat_set.insert((*num)/2).first;
+		multiplier_t fac=(*num)/2;
+		fac.canonicalize();
+		num=rat_set.insert(fac).first;
 		}
 
 	bool str_node::operator<(const cadabra::str_node& other) const
 		{
 		if(*name<*other.name) return true;
-		else return false;
+		if(*name==*other.name && *multiplier<*other.multiplier) return true;
+		return false;
 		}
 
 	}
@@ -1005,4 +1095,3 @@ std::ostream& operator<<(std::ostream& str, cadabra::Ex::iterator it)
 	//	ex.print_python(str, ex.begin());
 	return str;
 	}
-
