@@ -4,6 +4,7 @@
 #include "Functional.hh"
 #include "algorithms/canonicalise.hh"
 #include "modules/xperm_new.h"
+#include "properties/Trace.hh"
 #include "properties/Traceless.hh"
 #include "properties/Diagonal.hh"
 #include "properties/Derivative.hh"
@@ -22,15 +23,17 @@ canonicalise::canonicalise(const Kernel& k, Ex& tr)
 bool canonicalise::can_apply(iterator it)
 	{
 	if(*(it->name)!="\\prod")
-		if(is_single_term(it)==false)
+		if(*(it->name)=="\\pow" || is_single_term(it)==false)
 			return false;
 
+	//if(! (tr.is_head(it)==false && *tr.parent(it)->name=="\\pow" && tr.index(it)==0) )
+
 	// Canonicalise requires strict monomial structure: no products which contain
 	// sums as factors. Products as factors are ok, they do not lead to multiple
 	// identically named free indices.
 
-	auto sum_or_prod = find_in_subtree(tr, it, [](Ex::iterator tst) {
-		if(*tst->name=="\\sum") return true;
+	auto sum_or_prod = find_in_subtree(tr, it, [this](Ex::iterator tst) {
+		if(*tst->name=="\\sum" && number_of_indices(tst)>0) return true;
 		return false;
 		}, false);
 	if(sum_or_prod!=tr.end()) {
@@ -40,6 +43,9 @@ bool canonicalise::can_apply(iterator it)
 		return false;
 		}
 
+#ifdef DEBUG
+	std::cerr << "canonicalise::can_apply: at " << it << std::endl;
+#endif
 	return true;
 	}
 
@@ -48,20 +54,55 @@ bool canonicalise::remove_traceless_traces(iterator& it)
 	// Remove any traces of traceless tensors (this is best done early).
 	sibling_iterator facit=tr.begin(it);
 	while(facit!=tr.end(it)) {
-		const Traceless *trl=kernel.properties.get_composite<Traceless>(facit);
+		const Traceless *trl=kernel.properties.get<Traceless>(facit);
 		if(trl) {
+			unsigned int ihits=0;
 			tree_exact_less_mod_prel_obj comp(&kernel.properties);
 			std::set<Ex, tree_exact_less_mod_prel_obj> countmap(comp);
 			index_iterator indit=begin_index(facit);
 			while(indit!=end_index(facit)) {
+				bool incremented_now=false;
+				auto ind=kernel.properties.get<Indices>(indit, true);
+				if(ind) {
+					// The indexs need to be in the set for which the object is
+					// traceless (if specified, otherwise accept all).
+					if(trl->index_set_names.find(ind->set_name)!=trl->index_set_names.end() || trl->index_set_names.size()==0) {
+						incremented_now=true;
+						++ihits;
+						}
+					}
+				else incremented_now=true;
+				// Having no name is treated as having the right name
 				if(countmap.find(Ex(indit))==countmap.end()) {
 					countmap.insert(Ex(indit));
-					++indit;
 					}
-				else {
+				else if(incremented_now) {
 					zero(it->multiplier);
 					return true;
 					}
+				++indit;
+				}
+			iterator parent=it;
+			if (tr.number_of_children(it)==1 && !tr.is_head(it)) parent=tr.parent(it);
+			const Trace *trace=kernel.properties.get<Trace>(parent);
+			if(trace) {
+				int tmp;
+				auto impi=kernel.properties.get_with_pattern<ImplicitIndex>(facit, tmp, "");
+				if(impi.first->explicit_form.size()>0) {
+					// Does the explicit form have two more indices of the right type?
+					Ex::iterator eform=impi.first->explicit_form.begin();
+					unsigned int ehits=0;
+					indit=begin_index(eform);
+					while(indit!=end_index(eform)) {
+						auto ind=kernel.properties.get<Indices>(indit, true);
+						if(trl->index_set_names.find(ind->set_name)!=trl->index_set_names.end() && ind->set_name==trace->index_set_name) ++ehits;
+						if(ehits - ihits > 1) {
+							zero(it->multiplier);
+							return true;
+							}
+						++indit;
+						}
+					}
 				}
 			}
 		++facit;
@@ -74,7 +115,7 @@ bool canonicalise::remove_vanishing_numericals(iterator& it)
 	// Remove Diagonal objects with numerical indices which are not all the same.
 	sibling_iterator facit=tr.begin(it);
 	while(facit!=tr.end(it)) {
-		const Diagonal *dgl=kernel.properties.get_composite<Diagonal>(facit);
+		const Diagonal *dgl=kernel.properties.get<Diagonal>(facit);
 		if(dgl) {
 			index_iterator indit=begin_index(facit);
 			if(indit->is_rational()) {
@@ -104,19 +145,6 @@ Indices::position_t canonicalise::position_type(iterator it) const
 	return Indices::free;
 	}
 
-std::string canonicalise::get_index_set_name(iterator it) const
-	{
-	const Indices *ind=kernel.properties.get<Indices>(it, true);
-	if(ind) {
-		return ind->set_name;
-		// TODO: The logic was once as below, but it is no longer clear to
-		// me why that would ever make sense.
-		//		if(ind->parent_name!="") return ind->parent_name;
-		//		else                     return ind->set_name;
-		}
-	else return " undeclared";
-	}
-
 bool canonicalise::only_one_on_derivative(iterator i1, iterator i2) const
 	{
 	int num=0;
@@ -363,10 +391,13 @@ Algorithm::result_t canonicalise::apply(iterator& it)
 		sibling_iterator facit=tr.begin(it);
 		int curr_pos=0;
 		while(facit!=tr.end(it)) {
-			const TableauBase *tba=kernel.properties.get_composite<TableauBase>(facit);
+			const TableauBase *tba=kernel.properties.get<TableauBase>(facit);
 			// std::cerr << Ex(facit) << " has tableaubase " << tba << std::endl;
 			if(tba) {
 				unsigned int num_ind=number_of_indices(facit);
+#ifdef XPERM_DEBUG
+				std::cerr << "factor " << *facit->name << " has " << num_ind << " indices" << std::endl;
+#endif
 
 				// Add indices to the base. We used to add everything except the last one, but that
 				// seems to be the wrong thing to do after the XPERM -> XPERM_EXT upgrade (see Jose's email).
