@@ -11,6 +11,7 @@
 #include <sstream>
 #include <internal/unistd.h>
 #include <map>
+#include <regex>
 #include <glibmm/spawn.h>
 #include "process.hpp"
 
@@ -114,6 +115,11 @@ unsigned TeXEngine::TeXRequest::height() const
 	return height_;
 	}
 
+std::string TeXEngine::TeXRequest::latex() const
+	{
+	return latex_string;
+	}
+
 const std::vector<unsigned char>& TeXEngine::TeXRequest::image() const
 	{
 	return image_;
@@ -208,6 +214,11 @@ void TeXEngine::set_geometry(int horpix)
 	horizontal_pixels_=horpix;
 	}
 
+int TeXEngine::get_geometry() const
+	{
+	return horizontal_pixels_;
+	}
+
 void TeXEngine::set_font_size(int fontsize)
 	{
 	if(font_size_!=fontsize) {
@@ -218,9 +229,15 @@ void TeXEngine::set_font_size(int fontsize)
 			++reqit;
 			}
 		}
+	// std::cerr << "FONTSIZE " << fontsize << std::endl;
 	font_size_=fontsize;
 	}
 
+int TeXEngine::get_font_size() const
+	{
+	return font_size_;
+	}
+
 void TeXEngine::set_scale(double total_scale, double device_scale)
 	{
 	if(total_scale_!=total_scale && device_scale_!=device_scale) {
@@ -240,6 +257,11 @@ double TeXEngine::get_scale() const
 	return total_scale_;
 	}
 
+double TeXEngine::get_device_scale() const
+	{
+	return device_scale_;
+	}
+
 TeXEngine::TeXRequest::TeXRequest()
 	: needs_generating(true)
 	{
@@ -287,17 +309,36 @@ std::shared_ptr<TeXEngine::TeXRequest> TeXEngine::modify(std::shared_ptr<TeXRequ
 
 void TeXEngine::convert_all()
 	{
-	size_t need_generating=0;
-	auto it=requests.begin();
-	while(it!=requests.end()) {
-		if((*it)->needs_generating)
-			++need_generating;
-		++it;
+	size_t needs_generating=0;
+	for(const auto& req: requests) {
+		if(req->needs_generating)
+			++needs_generating;
 		}
 
-	if(need_generating!=0 && horizontal_pixels_!=0) {
-		//std::cerr << "cadabra-client: running TeX on " << need_generating << " requests" << std::endl;
-		convert_set(requests);
+	if(needs_generating!=0 && horizontal_pixels_!=0) {
+		std::string error_string="";
+		do {
+			try {
+				// std::cerr << "TeXEngine::convert_all: running TeX on " << needs_generating << " requests" << std::endl;
+				convert_set(requests);
+				needs_generating=0;
+				}
+			catch(TeXException& e) {
+				// std::cerr << "TeXEngine::convert_all: error happened." << std::endl;
+				error_string += e.what();
+				needs_generating=0;
+				for(const auto& req: requests)
+					if(req->needs_generating)
+						++needs_generating;
+				// Run again; the erroneous cell has had its `needs_generating` flag set
+				// to false and its `has_error` set to true, so we can safely feed this
+				// through again.
+				}
+			}
+		while(needs_generating>0);
+
+		if(error_string!="")
+			throw TeXException(error_string);
 		}
 	}
 
@@ -360,7 +401,7 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 	// Note: the number here has no effect on the size in pixels of the generated
 	// PDF. That is set with the -D parameter of dvipng.
 
-	const double horizontal_mm=horizontal_pixels_*(12.0/font_size_)/3.94/(total_scale_/device_scale_);
+	const double horizontal_mm=horizontal_pixels_*(12.0/font_size_)/4.4/(total_scale_/device_scale_);
 	//#ifdef DEBUG
 	//	std::cerr << "tex_it: font_size " << font_size << std::endl
 	//				 << "        pixels    " << horizontal_pixels_ << std::endl
@@ -449,7 +490,7 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 #ifdef __CYGWIN__
 	// MikTeX does not see /tmp, it needs \cygwin\tmp
 	nf="\\cygwin"+nf;
-	pcrecpp::RE("/").GlobalReplace("\\\\", &nf);
+	nf.replace(nf.begin(), nf.end(), '/', '\\');
 #endif
 
 	// Run LaTeX on the .tex file.
@@ -466,29 +507,51 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 	// The options below are chosen so they are compatible with MikTeX (https://docs.miktex.org/manual/pdftex.html),
 	// standard TeXLive on Linux and OSX. We use pdflatex because several latex distributions do not seem to have
 	// a 'latex' command anymore which supports all these options (e.g. MikTeX).
+	// For some reason, recent pdflatex versions no longer do anything with --halt-on-error, so we
+	// now also add the -interaction=nonstopmode (errorstopmode does not work either).
 
 	std::string latex_stdout, latex_stderr;
-	tpl::Process latex_proc("pdflatex --quiet --halt-on-error --output-format=dvi "+nf, "",
+	tpl::Process latex_proc("pdflatex --halt-on-error -interaction=nonstopmode --output-format=dvi "+nf, "",
 	[&](const char *bytes, size_t n) {
-		latex_stdout=std::string(bytes,n);
+		latex_stdout+=std::string(bytes,n);
 		},
 	[&](const char *bytes, size_t n) {
-		latex_stderr=std::string(bytes,n);
+		latex_stderr+=std::string(bytes,n);
 		}
 	                       );
 	auto latex_exit_status=latex_proc.get_exit_status();
 
-	erase_file(tmppath+".aux");
-	erase_file(tmppath+".log");
-	erase_file(tmppath+".out");
-
 	std::string err=handle_latex_errors(latex_stdout+latex_stderr, latex_exit_status);
 	setenv("TEXINPUTS", oldtexinputs.c_str(), 1);
 
 	if(err.size()>0) {
+		// std::cerr << "ERROR: " << latex_stderr << std::endl;
+		// std::cerr << "OUTPUT: " << latex_stdout << std::endl;
+		// std::cerr << "FILE: " << nf << std::endl;
+		std::regex  pat(R"(\[([0-9]*)\])");
+		std::smatch sm;
+		int last_error_free_page=0;
+		while(std::regex_search( latex_stdout, sm, pat)) {
+			// std::cerr << "PAGE: " << sm.str() << std::endl;
+			last_error_free_page = std::stoi( sm[1].str() );
+			latex_stdout = sm.suffix();
+			}
+		// std::cerr << "PAGE WITH ERROR: " << last_error_free_page+1 << std::endl;
+		// Label the error request so it can be ignored next time.
 		reqit=reqs.begin();
-		while(reqit!=reqs.end())
-			(*reqit++)->needs_generating=false;
+		int current_page=1;
+		while(reqit!=reqs.end()) {
+			if((*reqit)->needs_generating) {
+				if(current_page == last_error_free_page+1) {
+					// std::cerr << "CONTAINS ERROR: " << (*reqit)->latex_string << std::endl;
+					(*reqit)->has_error=true;
+					(*reqit)->needs_generating=false;
+					}
+				++current_page;
+				(*reqit)->image_.clear();
+				}
+			++reqit;
+			}
 		erase_file(tmppath+".dvi");
 		if(chdir(olddir)==-1)
 			throw TeXException(err+" (and cannot chdir back to original "+olddir+").");
@@ -499,6 +562,10 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 		throw TeXException(err);
 		}
 
+	erase_file(tmppath+".aux");
+	erase_file(tmppath+".log");
+	erase_file(tmppath+".out");
+
 	erase_file(tmppath+".tex");
 
 	// Convert the entire dvi file to png files.
@@ -528,7 +595,6 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 				pngname << tmppath << pagenum << ".png";
 				erase_file(pngname.str());
 				(*reqit)->image_.clear();
-				(*reqit)->needs_generating=true;
 				++pagenum;
 				}
 			++reqit;
@@ -569,4 +635,3 @@ void TeXEngine::convert_set(std::set<std::shared_ptr<TeXRequest> >& reqs)
 	if(chdir(olddir)==-1)
 		throw TeXException("Failed to chdir back to " +std::string(olddir)+".");
 	}
-
